package com.ksyun.service;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.text.SimpleDateFormat;
import java.util.*;

public class Request {
    public static final String VERSION = "1.0.0";

    public static final String SIGNATURE_VERSION = "1.0";
    public static final String ALGORITHM = "HMAC-SHA256";
    public static final String DATE_FORMAT = "yyyy-MM-dd'T'HH:mm:ss'Z'";
    public static final String DEFAULT_CONTENT_TYPE = "application/x-www-form-urlencoded";

    private String mAppKey;
    private String mAppSecret;
    private String mMethod;
    private String mUrl;

    public Map<String, String> mPathArgs;
    public Map<String, String> mQueryArgs;
    public Map<String, String> mBodyArgs;
    public Map<String, String> mHeaders;
    public String mBodyString;

    private String mNonceStr;
    private boolean mBodyIsForm;

    public Request(String method, String url) {
        mMethod = method;
        mUrl = url;
        mNonceStr = _nonceStr(16);
    }

    public Request(String method, String url, Map<String, String>pathArgs, Map<String, String>queryArgs, Map<String, String>bodyArgs, Map<String, String>headers) {
        mMethod = method;
        mUrl = url;
        mPathArgs = pathArgs;
        mQueryArgs = queryArgs;
        mBodyArgs = bodyArgs;
        mHeaders = headers;
        mNonceStr = _nonceStr(16);
    }

    /**
     * setAKSK
     * @param appKey appKey
     * @param appSecret appSecret
     */
    public void setAKSK(String appKey, String appSecret) {
        mAppKey = appKey;
        mAppSecret = appSecret;
    }

    /**
     * canonicalQueryString
     * @return string
     */
    public String canonicalQueryString() {
        Map<String, String>signArgs = mapMerge(mHeaders, mBodyIsForm ? mBodyArgs : null);
        signArgs = mapMerge(signArgs, mQueryArgs);
        signArgs = mapMerge(signArgs, mPathArgs);
        signArgs = mapMerge(signArgs, publicHeaders(false));

        Map<String, String> map = new TreeMap<String, String>(
                new Comparator<String>() {
                    public int compare(String o1, String o2) {
                        return o1.compareTo(o2);
                    }
                }
        );
        for (String key : signArgs.keySet()) {
            map.put(key, signArgs.get(key));
        }

        List<String> canonicalSignArgs = new ArrayList<String>();
        for (String key : map.keySet()) {
            canonicalSignArgs.add(_escapeString(key) + "=" + _escapeString(map.get(key)));
        }
        return listJoin(canonicalSignArgs, "&");
    }

    /**
     * signString
     * @return string
     */
    public String signString() {
        try {
            Mac mac = Mac.getInstance("HmacSHA256");
            SecretKeySpec secretKeySpec = new SecretKeySpec(mAppSecret.getBytes(), "HmacSHA256");
            mac.init(secretKeySpec);

            byte[] array = mac.doFinal(canonicalQueryString().getBytes("UTF-8"));
            StringBuilder sb = new StringBuilder();
            for (byte item : array) {
                sb.append(Integer.toHexString((item & 0xFF) | 0x100).substring(1, 3));
            }
            return sb.toString();
        } catch (Exception e) {
            System.out.println(e.getMessage());
            return "";
        }
    }

    /**
     * publicHeaders
     * @param withSignature withSignature
     * @return headers
     */
    private Map<String, String> publicHeaders(boolean withSignature) {
        Map<String, String>headers = new HashMap<String, String>();
        headers.put("x-kscapigw-apigwak", mAppKey);
        headers.put("x-kscapigw-timestamp", _utcTime());
        headers.put("x-kscapigw-signatureversion", SIGNATURE_VERSION);
        headers.put("x-kscapigw-signaturemethod", ALGORITHM);
        headers.put("x-kscapigw-nonce", mNonceStr);

        if (withSignature) {
            headers.put("x-kscapigw-signed-headers", signedHeaders(headers));
            headers.put("x-kscapigw-signature", signString());
        }

        return headers;
    }

    /**
     * signedHeaders
     * @param publicHeaders headers
     * @return signedHeaders
     */
    private String signedHeaders(Map<String, String>publicHeaders) {
        return listJoin(new ArrayList<String>(mapMerge(mHeaders, publicHeaders).keySet()), ",");
    }

    /**
     * RequestHeaders
     * @return requestHeaders
     */
    private Map<String, String> requestHeaders() {
        return mapMerge(mHeaders, publicHeaders(true));
    }

    /**
     * mapMerge
     * @param map1 map
     * @param map2 map
     * @return map
     */
    private Map<String, String> mapMerge(Map<String, String>map1, Map<String, String>map2) {
        Map<String, String> map = new HashMap<String, String>();
        if (map1 != null) {
            for (String key : map1.keySet()) {
                map.put(key, map1.get(key));
            }
        }
        if (map2 != null) {
            for (String key : map2.keySet()) {
                map.put(key, map2.get(key));
            }
        }
        return map;
    }

    /**
     * Set Join
     * @param list list
     * @param sep sep
     * @return joined string
     */
    private String listJoin(List<String>list, String sep) {
        StringBuffer sb = new StringBuffer();
        for (String key : list) {
            sb.append(key + sep);
        }
        return sb.deleteCharAt(sb.length()-1).toString();
    }

    /**
     * buildQueryArgs
     * @param args map
     * @return string
     */
    private String buildQueryArgs(Map<String, String>args) {
        StringBuffer sb = new StringBuffer();
        for (String key : args.keySet()) {
            sb.append(key + "=" + args.get(key) + "&");
        }
        return sb.length() > 0 ? sb.deleteCharAt(sb.length()-1).toString() : "";
    }

    /**
     * escapeString
     * @param str str
     * @return escapeString
     */
    private String _escapeString(String str) {
        try {
            String es = URLEncoder.encode(str, "UTF-8");
            es = es.replace("+", "%20");
            es = es.replace("*", "%2A");
            es = es.replace("%7E", "~");
            return es;
        } catch (Exception e) {
            return str;
        }
    }

    /**
     * utcTime
     * @return string
     */
    private String _utcTime() {
        SimpleDateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);
        dateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
        return dateFormat.format(new Date());
    }

    /**
     * RandString
     * @param length int
     * @return string
     */
    private String _nonceStr(int length) {
        String str = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
        int strLen = str.length();

        Random random = new Random();
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < length; i++) {
            int index = random.nextInt(strLen);
            sb.append(str.charAt(index));
        }
        return sb.toString();
    }

    /**
     * CheckContentType
     */
    private void _checkContentType() {
        if (!mHeaders.containsKey("content-type")) {
            mHeaders.put("content-type", DEFAULT_CONTENT_TYPE);
        }
        mBodyIsForm = mHeaders.get("content-type").equals(DEFAULT_CONTENT_TYPE);
    }

    /**
     * execute
     */
    public String execute() {
        _checkContentType();

        HttpURLConnection connection = null;
        String result = "";

        try {
            if (!mQueryArgs.isEmpty()) {
                mUrl += "?" + buildQueryArgs(mQueryArgs);
            }

            URL url = new URL(mUrl);
            connection = (HttpURLConnection) url.openConnection();
            connection.setRequestMethod(mMethod);
            connection.setConnectTimeout(3000);
            connection.setDoInput(true);
            connection.setDoOutput(true);
            connection.setUseCaches(false);
            connection.setInstanceFollowRedirects(false);

            Map<String, String>headers = requestHeaders();
            for (String key : headers.keySet()) {
                connection.setRequestProperty(key, headers.get(key));
            }

            String params = "";
            String contentType = mHeaders.get("content-type");
            if (contentType.equals(DEFAULT_CONTENT_TYPE)) {
                params = buildQueryArgs(mBodyArgs);
            } else {
                params = mBodyString;
            }

            if (params.length() > 0) {
                DataOutputStream out = new DataOutputStream(connection.getOutputStream());
                out.write(params.getBytes());
                out.flush();
                out.close();
            }

            connection.connect();

            // read
            BufferedReader reader = new BufferedReader(new InputStreamReader(connection.getResponseCode() == 200 ? connection.getInputStream() : connection.getErrorStream()));
            String line;
            while ((line = reader.readLine()) != null) {
                result += line + "\n";
            }
            reader.close();
        } catch (IOException e) {
            System.out.println("err:" + e.getLocalizedMessage());
        } finally {
            if (connection != null) {
                connection.disconnect();
            }
        }
        return result;
    }
}
